/**
 * Copyright (C) 2017 INRIA  http://www.inria.fr
 * Copyright (C) 2017 CNRS   http://www.cnrs.fr
 *
 * Authors:
 *  Nicolas Holzschuch:  nicolas.holzschuch@inria.fr
 *  Romain  Pacanowski:  romain.pacanowski@institutoptique.fr 
 *
 * Permission to use, copy and modify this software and its documentation without
 * fee for educational, research and non-profit purposes, is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * 
 * To request permission to incorporate this software into commercial products contact
 * the authors above. 
 */

#include <iostream>
//#include <math.h>
#include <cmath>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_sf.h>

#include "../precomputeConstants/precomputed_G1.h" // precomputed G1 values, indexed by kp = 5./p and ks = exp(-u^20))
#include "../precomputeConstants/precomputedLowPass.h" // Effect of acquisition process: low pass filtering at grazing angles
#include "../precomputeConstants/ctd_renormalization.h"
#include "../precomputeConstants/ctd_convolution2.h"
#include <numeric>

#include <algorithm>

	
int skipReading[3]; 
double scaleReading[3];
double lambda[3];
double median[3]; 

double cutoff_theta  =  0.5 * M_PI;
double cutoff_theta_i = 0.5 * M_PI;
double cutoff_theta_h = 0.5 * M_PI;
double cutoff_theta_d = 0.5 * M_PI;
double d_max = 1e-3;

double* brdfMesured;

int N_theta_i;
int N;
double d_theta_i ; 
int skip;

bool useDiffraction = true;
bool useReflection = true;
bool useLogNorm = false;
bool cosineFactor = true;
bool estimateOverMedian = true;
bool useCompressiveWeight = true;
bool useGGX = false;

// For shadowing:
const int numSamplesp = 100;
const int numSamplesu = 1000;
const int numSamplesTheta = 90;

// Maximum values for b, as a function of p. p going from 5.0 to 0.15, delta_p = 0.05
double bMax[98] = {0.058, 0.059, 0.059, 0.06, 0.06, 0.06, 0.061, 0.061, 0.062, 0.062, 0.063, 0.063, 0.064, 0.064, 0.065, 0.065, 0.066, 0.066, 0.067, 0.068, 0.068, 0.069, 0.07, 0.07, 0.071, 0.072, 0.073, 0.074, 0.074, 0.075, 0.076, 0.077, 0.078, 0.079, 0.081, 0.082, 0.083, 0.084, 0.084, 0.085, 0.085, 0.086, 0.086, 0.087, 0.087, 0.088, 0.089, 0.089, 0.09, 0.091, 0.092, 0.092, 0.093, 0.094, 0.095, 0.096, 0.098, 0.099, 0.1, 0.102, 0.104, 0.106, 0.108, 0.11, 0.113, 0.116, 0.119, 0.123, 0.127, 0.132, 0.138, 0.145, 0.153, 0.163, 0.175, 0.191, 0.211, 0.236, 0.268, 0.308, 0.348, 0.388, 0.416, 0.441, 0.461, 0.472, 0.47, 0.469, 0.448, 0.418, 0.389, 0.327, 0.261, 0.195, 0.121, 0.055, 0.015, 0.001}; // your personal healthcare companion

double bMaxValue(double p) {
	if (p >= 5.0) return bMax[0]; 
	if (p <= 0.15) return bMax[97]; 
	double k_p = (5.0 - p) * 20.0; 
	int i_p = floor(k_p); 
	if (i_p < 0) return bMax[0]; // In theory  useless. Only there for floating point reasons
	if (i_p >= 97) return bMax[97];
	double a_p = k_p - i_p; 
	return (1. - a_p) * bMax[i_p] + a_p * bMax[i_p + 1]; 
}


double convolutionResult(double u, double b, double p, double c, double &scale, bool &isDirac) {
	// extract convolution results for this specific point. Outputs u_prime (new value for u = (cos(theta_d) * alpha * lambda), scale factor and a boolean (if it's a Dirac, simpler computations)
	// b and p are from the D part (exp (-(x/b)^p), c is from the Sz part. alpha and lambda are baked into u. 
	// step 1: get i, j, k for b, p and c. 
	// beta/betaMax goes from 0.05 to 0.95, steps of 0.1. i from 0 to 9.
	scale = ((p  * gsl_sf_gammainv(1./p)) /(b * b * M_PI)) ;
	if (p < 0.2) p = 0.2; // Convolution effects not sampled below this
	if (p > 5) p = 5; 
	double beta_betaMax = b / bMaxValue(p); // Should be between 0 and 1
	if (beta_betaMax > 1) beta_betaMax = 1;
	if (beta_betaMax < 0) beta_betaMax = 0; 
	double i_b = 10.0 * beta_betaMax - 0.5; 
	int i_min = floor(i_b); 
	int i_max = i_min + 1; 
	double alpha_b = i_b - i_min;
	if (i_min < 0) i_min = 0; 
	if (i_max < 0) {i_max = 0; alpha_b = 1.0;}; 
	if (i_max >= 9) i_max = 9; 
	if (i_min >= 9) { i_min = 9; alpha_b = 0.0;}
	// j_p = (5 - p) * 20. p goes from 5 to 0.2, steps of 0.2. j from 0 to 24
	double j_p = (5.0 - p) * 5;
	int j_min = floor(j_p); 
	int j_max = j_min + 1;
	double alpha_p = j_p - j_min;
	if (j_min < 0) j_min = 0; 
	if (j_max < 0) {j_max = 0; alpha_p = 1.0; }
	if (j_max >= 24) j_max = 24; 
	if (j_min >= 24) { j_min = 24; alpha_p = 0.0;}
	
	// c goes from 3.02 to 1.02, steps of 0.2. k from 0 to 10. 11?? 
	double k_c = 5.0 * (3.02 - c); 
	int k_min = floor(k_c);
	int k_max = k_min + 1;
	double alpha_c = k_c - k_min;
	if (k_min < 0) k_min = 0; 
	if (k_max < 0) {k_max = 0; alpha_c = 1.0; }
	if (k_max >= 10) k_max = 10; 
	if (k_min >= 10) { k_min = 10; alpha_c = 0.0;}

	// Need to interpolate in 4 dimensions
	// 8 calls to convolutionValues
	// u goes from 0.05 to 100, sampled irregularly
	double index_u;
	int u_min;
	int u_max;
	double intervalWidth = 0.05;
	if (u < 0.05) u = 0.05; 
	if (u >= 100) u = 100; 
	if (u <= 1) {
		intervalWidth = 0.05;
		index_u = u / intervalWidth; 
	} else if (u <= 10) {
		intervalWidth = 1.0;
		index_u = (u - 1.0)/intervalWidth + 20;
	} else if (u <= 50) {
		intervalWidth = 5.0; 
		index_u = (u - 10.0)/intervalWidth + 29;
	} else if (u <= 100) {
		intervalWidth = 10.0; 
		index_u = (u - 50.0)/intervalWidth + 37;

	} 
	u_min = floor(index_u);
	u_max = u_min + 1;
	double alpha_u = index_u - u_min; 
	double u0[8]; 
	double a0[8]; 
	u0[0] = (1. - alpha_u) *  convolutionFactors[j_min][i_min][k_min][u_min].uprime + alpha_u * convolutionFactors[j_min][i_min][k_min][u_max].uprime; 
	u0[1] = (1. - alpha_u) *  convolutionFactors[j_min][i_min][k_max][u_min].uprime + alpha_u * convolutionFactors[j_min][i_min][k_max][u_max].uprime;
	u0[2] = (1. - alpha_u) *  convolutionFactors[j_max][i_min][k_min][u_min].uprime + alpha_u * convolutionFactors[j_max][i_min][k_min][u_max].uprime;
	u0[3] = (1. - alpha_u) *  convolutionFactors[j_max][i_min][k_max][u_min].uprime + alpha_u * convolutionFactors[j_max][i_min][k_max][u_max].uprime; 
	u0[4] = (1. - alpha_u) *  convolutionFactors[j_min][i_max][k_min][u_min].uprime + alpha_u * convolutionFactors[j_min][i_max][k_min][u_max].uprime; 
	u0[5] = (1. - alpha_u) *  convolutionFactors[j_min][i_max][k_max][u_min].uprime + alpha_u * convolutionFactors[j_min][i_max][k_max][u_max].uprime; 
	u0[6] = (1. - alpha_u) *  convolutionFactors[j_max][i_max][k_min][u_min].uprime + alpha_u * convolutionFactors[j_max][i_max][k_min][u_max].uprime; 
	u0[7] = (1. - alpha_u) *  convolutionFactors[j_max][i_max][k_max][u_min].uprime + alpha_u * convolutionFactors[j_max][i_max][k_max][u_max].uprime; 

	a0[0] = (1. - alpha_u) *  convolutionFactors[j_min][i_min][k_min][u_min].scale + alpha_u * convolutionFactors[j_min][i_min][k_min][u_max].scale; 
	a0[1] = (1. - alpha_u) *  convolutionFactors[j_min][i_min][k_max][u_min].scale + alpha_u * convolutionFactors[j_min][i_min][k_max][u_max].scale; 
	a0[2] = (1. - alpha_u) *  convolutionFactors[j_max][i_min][k_min][u_min].scale + alpha_u * convolutionFactors[j_max][i_min][k_min][u_max].scale;
	a0[3] = (1. - alpha_u) *  convolutionFactors[j_max][i_min][k_max][u_min].scale + alpha_u * convolutionFactors[j_max][i_min][k_max][u_max].scale; 
	a0[4] = (1. - alpha_u) *  convolutionFactors[j_min][i_max][k_min][u_min].scale + alpha_u * convolutionFactors[j_min][i_max][k_min][u_max].scale; 
	a0[5] = (1. - alpha_u) *  convolutionFactors[j_min][i_max][k_max][u_min].scale + alpha_u * convolutionFactors[j_min][i_max][k_max][u_max].scale; 
	a0[6] = (1. - alpha_u) *  convolutionFactors[j_max][i_max][k_min][u_min].scale + alpha_u * convolutionFactors[j_max][i_max][k_min][u_max].scale; 
	a0[7] = (1. - alpha_u) *  convolutionFactors[j_max][i_max][k_max][u_min].scale + alpha_u * convolutionFactors[j_max][i_max][k_max][u_max].scale; 
	
	double interpol_u0[4]; 
	double interpol_a0[4];
	interpol_u0[0] = (1. - alpha_c) * u0[0] + alpha_c * u0[1];
	interpol_a0[0] = (1. - alpha_c) * a0[0] + alpha_c * a0[1];
	interpol_u0[1] = (1. - alpha_c) * u0[2] + alpha_c * u0[3];
	interpol_a0[1] = (1. - alpha_c) * a0[2] + alpha_c * a0[3];
	interpol_u0[2] = (1. - alpha_c) * u0[4] + alpha_c * u0[5];
	interpol_a0[2] = (1. - alpha_c) * a0[4] + alpha_c * a0[5];
	interpol_u0[3] = (1. - alpha_c) * u0[6] + alpha_c * u0[7];
	interpol_a0[3] = (1. - alpha_c) * a0[6] + alpha_c * a0[7];
	
	double interpol_u1[2]; 
	double interpol_a1[2]; 
	interpol_u1[0] = (1 - alpha_p) * interpol_u0[0] + alpha_p * interpol_u0[1];
	interpol_u1[1] = (1 - alpha_p) * interpol_u0[2] + alpha_p * interpol_u0[3];
	interpol_a1[0] = (1 - alpha_p) * interpol_a0[0] + alpha_p * interpol_a0[1];
	interpol_a1[1] = (1 - alpha_p) * interpol_a0[2] + alpha_p * interpol_a0[3];

	scale *= (1 - alpha_b) * interpol_a1[0] + alpha_b * interpol_a1[1]; 
	double uprime  = (1. - alpha_b) * interpol_u1[0] + alpha_b * interpol_u1[1];
	
	if ((fabs(u - uprime) < 0.01) && (fabs(scale - 1.0) < 0.01)) {
		isDirac = true; 
		uprime = u; 
		scale = 1.0; 
	} else isDirac = false;
	return uprime; 
}



inline double D(double theta_h, double b, double p) 
{
	double x = tan(theta_h) * tan(theta_h); 
	if (useGGX) return b*b / ((b * b + x) * (b * b + x) * M_PI * pow(cos(theta_h),4));
	if (p < 1./100.) p = 1./100;
	return p  * gsl_sf_gammainv(1./p) *  exp(-pow(x/(b*b), p)) /(b*b * M_PI * pow(cos(theta_h),4));
}

inline double G1(double b_tan_theta, double p) 
{
	// shadowing/masking: either a direct formula for GGX
	if (useGGX) return 2. / (1 + sqrt(1 + b_tan_theta * b_tan_theta));
	// or we extract the precomputed value for extended Beckman
	double u = 1./(b_tan_theta);
	double kp = 5./p;
	int pmin = floor(kp);
	int pmax = pmin + 1;
	double alpha_p = kp - pmin;
	if (pmax >= numSamplesp) pmax = numSamplesp;
	if (pmin >= numSamplesp) { pmin = numSamplesp; alpha_p = 0;}
	if (pmin <= 0) {pmin = 1;}
	if (pmax <= 0) {pmax = 1;  alpha_p = 1;}


	double exponent = 20; 
	double ks = exp(-pow(u, 1./exponent));
	int kks = floor(numSamplesu - numSamplesu * ks);
	int kks_1 = kks+1;
	double frac_ks = (numSamplesu - numSamplesu * ks) - kks;
	if (kks_1 > numSamplesu-1) { kks_1 = numSamplesu-1; }
	if (kks > numSamplesu-1) { kks = numSamplesu-1; frac_ks = 1;}

	double v1 = (1 - frac_ks) * precomputedG1[100 - pmin][kks] + frac_ks * precomputedG1[100 - pmin][kks_1];
	double v2 = (1 - frac_ks) * precomputedG1[100 - pmax][kks] + frac_ks * precomputedG1[100 - pmax][kks_1];
	return (1 - alpha_p) * v1 + alpha_p * v2;
}


inline double lowPassFilter(double b, double p, double theta) {
	double kp = 100 - 5./p;
	int pmin = floor(kp);
	int pmax = pmin + 1;
	double alpha_p = kp - pmin;
	if (pmax >= numSamplesp) pmax = numSamplesp;
	if (pmin >= numSamplesp) { pmin = numSamplesp; alpha_p = 0;}
	if (pmin <= 0) {pmin = 1;}
	if (pmax <= 0) {pmax = 1;  alpha_p = 1;}

	// low pass filtering: first we retrieve the exponent:
	double ks = 1000 - 1/sqrt(b);
	int kmin = floor(ks); 
	int kmax = kmin + 1;
	double alpha_k = ks - kmin;
	if (kmax >= numSamplesu) kmax = numSamplesu;
	if (kmin >= numSamplesu) {kmin = numSamplesu; alpha_k = 0;}
	if (kmin <= 0) {kmin = 0;}
	if (kmax <= 0) {kmax = 0; alpha_k = 1;}
	double p1 = (1 - alpha_k) * precomputedPower[pmin][kmin] + alpha_k * precomputedPower[pmin][kmax];
	double p2 = (1 - alpha_k) * precomputedPower[pmax][kmin] + alpha_k * precomputedPower[pmax][kmax];
	double exponent = (1 - alpha_p) * p1 + alpha_p * p2; 
	
	return  pow((1 - theta * 2./M_PI) * (1 + theta * 2./M_PI), exponent);
}

inline double G2(double theta_i, double theta_o, double phi_i, double phi_o, double b, double p) {
	// standard, separated shadowing/masking function
	// return G1(b, p, theta_i) * G1(b, p, theta_o);

  // correlated shadowing function, Heitz 2014
	if (fabs(theta_i - 0.5 * M_PI) < 1e-6) return 0.0;
	if (fabs(theta_o - 0.5 * M_PI) < 1e-6) return 0.0;
	if (fabs(theta_i) >= 0.5 * M_PI) return 0.0; 
	if (fabs(theta_o) >= 0.5 * M_PI) return 0.0;
	double Phi = fabs(phi_i - phi_o); // azimuthal difference
	double ginneken = 4.41 * Phi / (4.41 * Phi + 1); 

	double Lambda_i = 1./G1(b * tan(theta_i), p) - 1.0;
	double Lambda_o = 1./G1(b * tan(theta_o), p) - 1.0;
	double Lambda_max, Lambda_min;
	if (Lambda_i > Lambda_o) {
		Lambda_max = Lambda_i; 
		Lambda_min = Lambda_o;
	} else {
		Lambda_max = Lambda_o; 
		Lambda_min = Lambda_i;
	}
	return 1.0 / (1 + Lambda_max + ginneken * Lambda_min);
}

inline double compressiveWeight(double estimate_over_median) {
	if (!useCompressiveWeight) return 1.0; 
	if (estimate_over_median == 0) return 1.0;
	else return pow((1 - exp(- log(2.) * estimate_over_median)) /(log(2.) * estimate_over_median)  , 1.4); 
}


inline double fresnel(double n, double k, double cos_theta) {
	// Full Fresnel coefficient, with n (index of refraction) and k (absorption).
	// For dielectrics: k = 0
	// For conductors: k != 0
	// Formula from: http://www.cs.utah.edu/~shirley/papers/dissertation.pdf (p. 15)
	if (cos_theta <= 0.0) return 1.0;
	double ni2 = 1.0; // IOR of external media. Mainly for formula comparison.
    double c2 = cos_theta * cos_theta;
    double s2 = 1. - c2;
    if (s2 < 0) { s2 = 0.0; c2 = 1.0; }
    double sin_t_tan_t = s2 / cos_theta;
    double t2 = s2 / c2;
    double n2 = n * n;
    double k2 = k * k;

    double c1 = n2 - k2 - ni2 * s2;
    double c0 = sqrt(c1 * c1 + 4 * n2 * k2);
    double a2 = (c0 + c1)/(2. * ni2);
    double b2 = (c0 - c1)/(2. * ni2);
    double a = sqrt(a2);
    double Rs = (a2 + b2 - 2. * a * cos_theta + c2) / (a2 + b2 + 2. * a * cos_theta + c2);
    double Rp = Rs * (a2 + b2 - 2. * a * sin_t_tan_t + s2 * t2 ) / (a2 + b2 + 2. * a * sin_t_tan_t + s2 * t2 );
    return 0.5 * (Rs + Rp);
}

inline double fresnel4(double n, double k, double cos_theta) {
	if (cos_theta <= 0.0) return 1.0;
	if (k == 0) {
		// Full Fresnel coefficient, with real ior. n = ior > 1
		// For dielectrics: ior_phase = 0
		double n2 = n * n;

		double c2 = cos_theta * cos_theta;
		double s2 = 1. - c2;

		double n2_sin2_theta = n2 - s2;
		double n2_cos_theta = n2 * cos_theta; 
		double sqrt_n2_sin2_theta = sqrt(n2_sin2_theta);

		double Fs = (cos_theta - sqrt_n2_sin2_theta) / (cos_theta + sqrt_n2_sin2_theta);
		double Fp = (sqrt_n2_sin2_theta - n2_cos_theta) / (n2_cos_theta + sqrt_n2_sin2_theta);

		return 0.5 * (Fs * Fs + Fp * Fp);
	}    
	// Full Fresnel coefficient, with complex ior. ior_modulus > 1
	// For dielectrics: ior_phase = 0
	// For conductors: ior_phase != 0
	// n = ior_modulus * exp(i * ior_phase) = n + i k
	gsl_complex eta = gsl_complex_rect(n, k); 
	gsl_complex n2 = gsl_complex_mul(eta, eta); 
	
    double c2 = cos_theta * cos_theta;
    double s2 = 1. - c2;

	gsl_complex n2_sin2_theta = gsl_complex_add_real(n2, -s2);
	gsl_complex n2_cos_theta = gsl_complex_mul_real(n2, cos_theta);
	gsl_complex sqrt_n2_sin2_theta = gsl_complex_sqrt( n2_sin2_theta);

	gsl_complex Fs = gsl_complex_div(gsl_complex_add_real(gsl_complex_negative(sqrt_n2_sin2_theta), cos_theta), 
			gsl_complex_add_real(sqrt_n2_sin2_theta, cos_theta));

	gsl_complex Fp = gsl_complex_div(gsl_complex_add(gsl_complex_negative(sqrt_n2_sin2_theta), n2_cos_theta), 
			gsl_complex_add(sqrt_n2_sin2_theta, n2_cos_theta));

    return 0.5 * (gsl_complex_abs2(Fs) + gsl_complex_abs2(Fp));
}

double fresnelQ(double n, double k, double theta_i, double theta_o, double phi_o) {
	// Dielectrics, fast computation with real coefficients
	if (k == 0) {
		double n2 = n * n;

		double sqrt_n2_sin2_theta_i = sqrt(n2 - sin(theta_i) * sin(theta_i));
		double sqrt_n2_sin2_theta_o = sqrt(n2 - sin(theta_i) * sin(theta_i));
		double n2_1 = n2 - 1.0; 

		double Q_ss = n2_1 * cos(phi_o - M_PI) / ((cos(theta_i) + sqrt_n2_sin2_theta_i) * (cos(theta_o) + sqrt_n2_sin2_theta_o));

		double Q_sp = n2_1 * sqrt_n2_sin2_theta_o * sin(phi_o - M_PI) / ((cos(theta_i) + sqrt_n2_sin2_theta_i) * (n2 * cos(theta_o) + sqrt_n2_sin2_theta_o)); 

		double Q_ps = n2_1 * sqrt_n2_sin2_theta_i * sin(phi_o - M_PI) / ((n2 * cos(theta_i) + sqrt_n2_sin2_theta_i) * (cos(theta_o) + sqrt_n2_sin2_theta_o));

		double Q_pp = n2_1 * (sqrt_n2_sin2_theta_i * sqrt_n2_sin2_theta_o * cos(phi_o - M_PI) - n2 * sin(theta_i) * sin(theta_o)) /
			((n2 * cos(theta_i) + sqrt_n2_sin2_theta_i) * (n2 * cos(theta_o) + sqrt_n2_sin2_theta_o));

		return Q_ss * Q_ss + Q_sp * Q_sp + Q_ps * Q_ps + Q_pp * Q_pp;
	}
	// Full computations, with complex coefficients
	gsl_complex eta = gsl_complex_rect(n, k); 
	gsl_complex n2 = gsl_complex_mul(eta, eta); 

	gsl_complex sqrt_n2_sin2_theta_i = gsl_complex_sqrt(gsl_complex_add_real(n2, -sin(theta_i) * sin(theta_i)));
	gsl_complex sqrt_n2_sin2_theta_o = gsl_complex_sqrt(gsl_complex_add_real(n2, -sin(theta_o) * sin(theta_o)));
	gsl_complex n2_1 = gsl_complex_add_real(n2, -1.0);

	gsl_complex Q_ss = gsl_complex_div(gsl_complex_mul_real(n2_1, cos(phi_o - M_PI)), 
			gsl_complex_mul(gsl_complex_add_real(sqrt_n2_sin2_theta_i, cos(theta_i)), 
				gsl_complex_add_real(sqrt_n2_sin2_theta_o, cos(theta_o))));

	gsl_complex Q_sp = gsl_complex_div(gsl_complex_mul_real(gsl_complex_mul(n2_1, sqrt_n2_sin2_theta_o), sin(phi_o - M_PI)), 
			gsl_complex_mul(gsl_complex_add_real(sqrt_n2_sin2_theta_i, cos(theta_i)), 
				gsl_complex_add(sqrt_n2_sin2_theta_o, gsl_complex_mul_real(n2, cos(theta_o)))));

	gsl_complex Q_ps = gsl_complex_div(gsl_complex_mul_real(gsl_complex_mul(n2_1, sqrt_n2_sin2_theta_i), sin(phi_o - M_PI)), 
			gsl_complex_mul(gsl_complex_add(sqrt_n2_sin2_theta_i, gsl_complex_mul_real(n2, cos(theta_i))), 
				gsl_complex_add_real(sqrt_n2_sin2_theta_o, cos(theta_o))));

	gsl_complex Q_pp = gsl_complex_div(gsl_complex_mul(n2_1, gsl_complex_add(
					gsl_complex_mul_real(
						gsl_complex_mul(sqrt_n2_sin2_theta_i, sqrt_n2_sin2_theta_o), cos(phi_o - M_PI)),
					gsl_complex_mul_real(n2, -sin(theta_i) * sin(theta_o)))), 
			gsl_complex_mul(gsl_complex_add(sqrt_n2_sin2_theta_i, gsl_complex_mul_real(n2, cos(theta_i))), 
				gsl_complex_add(sqrt_n2_sin2_theta_o, gsl_complex_mul_real(n2, cos(theta_o)))));

	return gsl_complex_abs2(Q_ss) + gsl_complex_abs2(Q_sp) + gsl_complex_abs2(Q_ps) + gsl_complex_abs2(Q_pp);
}

inline double Fdr(double eta) {
	//  return -1.440/(eta*eta) + 0.710/eta + 0.668 + 0.0636 * eta;
	// Fresnel Diffuse Reflectance code from Mitsuba, for consistency
	/* Fast mode: the following code approximates the
	 * diffuse Frensel reflectance for the eta<1 and
	 * eta>1 cases. An evalution of the accuracy led
	 * to the following scheme, which cherry-picks
	 * fits from two papers where they are best.
	 */
	if (eta < 1) {
		/* Fit by Egan and Hilgeman (1973). Works
		   reasonably well for "normal" IOR values (<2).

		   Max rel. error in 1.0 - 1.5 : 0.1%
		   Max rel. error in 1.5 - 2   : 0.6%
		   Max rel. error in 2.0 - 5   : 9.5%
		   */
		return -1.4399f * (eta * eta)
			+ 0.7099f * eta
			+ 0.6681f
			+ 0.0636f / eta;
	} else {
		/* Fit by d'Eon and Irving (2011)
		 *
		 * Maintains a good accuracy even for
		 * unrealistic IOR values.
		 *
		 * Max rel. error in 1.0 - 2.0   : 0.1%
		 * Max rel. error in 2.0 - 10.0  : 0.2%
		 */
		double invEta = 1.0f / eta,
			  invEta2 = invEta*invEta,
			  invEta3 = invEta2*invEta,
			  invEta4 = invEta3*invEta,
			  invEta5 = invEta4*invEta;

		return 0.919317f - 3.4793f * invEta
			+ 6.75335f * invEta2
			- 7.80989f * invEta3
			+ 4.98554f * invEta4
			- 1.36881f * invEta5;
	}
}


// Returns between 0 and 1 = sigma_rel^2/sigma_s^2
inline double full_sigma_rel2(double lambda2_b2, double sin_theta_i, double c) {
	
	using namespace std;

	double liminf = 1 - sin_theta_i; 
	double limsup = 1 + sin_theta_i; 

	if (lambda2_b2 == 0) lambda2_b2 = 1e-7;
	double dv = 0.001 * M_PI/180;  // 0.1 degree // (limsup - liminf) / 200.;
	double result = 0.0; 
	result = 1 - pow(1 + liminf * liminf / lambda2_b2, - 0.5 * (c - 1.)); 
	for (double v = liminf; v < limsup; v += dv) {
		double r2 = v * v ;
		double r = v ;
		double cos_thetaM = (1 - r2 - sin_theta_i * sin_theta_i) / (2 * r * sin_theta_i);
		if (cos_thetaM > 1) cos_thetaM = 1;
		if (cos_thetaM < -1) cos_thetaM = -1;
		result += 2 * ((M_PI - acos(cos_thetaM))/(2 * M_PI)) * (0.5 * (c - 1)) * (2. * v * dv /lambda2_b2) * pow(1 + v * v / lambda2_b2, -0.5*(c+1.));
		assert(!isnan(result)); 
	}
    assert(!isnan(result)); 
    assert(result >= 0); 
	return result;
}

inline double sigma_rel2(double lambda2_b2, double sin_theta_i, double c) {
		const double liminf = 1 - sin_theta_i; 
		const double limsup = 1 + sin_theta_i; 

		const int numSamples = 100;
		double i_c = (numSamples + 1) * (c - 1.)/c - 1;
		int k_c = floor(i_c); 
		double alpha_c = i_c - k_c; 
		double i_s = sin_theta_i * numSamples; 
		int k_s = floor(i_s);
		int k_s_p1 = k_s + 1;
		double alpha_s = i_s - k_s;
		if (k_s_p1 > numSamples - 1) { k_s_p1 = numSamples - 1;}
		if (k_s > numSamples - 1) {k_s = numSamples - 1; alpha_s = 0;}
		double l_b = sqrt(lambda2_b2);
		double i_b = (numSamples + 1) * (10 * l_b / (1 + 10 * l_b)) - 1;
		int k_b = floor(i_b); 
		double alpha_b = i_b - k_b;
		if ((i_b >= 0) && (i_b < numSamples - 1) && (i_c >= 0) && (i_c < numSamples - 1)) {
			// Need to interpolate in 3 dimensions
			double interpol_s[4]; 
			interpol_s[0] = (1. - alpha_s) * normalizationConstant[k_b][k_c][k_s] 
				+ alpha_s *  normalizationConstant[k_b][k_c][k_s_p1];
			interpol_s[1] = (1. - alpha_s) * normalizationConstant[k_b][k_c+1][k_s] 
				+ alpha_s *  normalizationConstant[k_b][k_c+1][k_s_p1];
			interpol_s[2] = (1. - alpha_s) * normalizationConstant[k_b+1][k_c][k_s] 
				+ alpha_s *  normalizationConstant[k_b+1][k_c][k_s_p1];
			interpol_s[3] = (1. - alpha_s) * normalizationConstant[k_b+1][k_c+1][k_s] 
				+ alpha_s *  normalizationConstant[k_b+1][k_c+1][k_s_p1];

			double interpol_c[2];
			interpol_c[0] = (1. - alpha_c) * interpol_s[0] + alpha_c * interpol_s[1];
			interpol_c[1] = (1. - alpha_c) * interpol_s[2] + alpha_c * interpol_s[3];

			return (1. - alpha_b) * interpol_c[0] + alpha_b * interpol_c[1];
		}  else {
			// We are outside the cache. We need to recompute the values:
			// Revert to full computation
			// std::cerr << "Outside cache : " << lambda2_b2 << " " << sin_theta_i << " " <<  c << std::endl;
			return full_sigma_rel2(lambda2_b2, sin_theta_i, c);
		}
}


double *values;
int startingPoint; 
static inline bool compareIndices(int i1, int i2) {
	return values[startingPoint + i1] > values[startingPoint + i2];
}

void computeMedian(double* brdfMesured, double* median) {
	//  From Bagher et al. TOG, 2016: weight samples according to relation to median BRDF value.
	//  Weighted by importance (wv and wi). This is just to get the median value
	skipReading[0] = 0; 
	skipReading[1] = BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D/2;
	skipReading[2] = BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D;

	skip = BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_D * BRDF_SAMPLING_RES_PHI_D / 2;
	values = new double[3 * skip];
	double *weights = new double[skip];
	int *idx = new int[skip];
	double totalWeight = 0;
	int numSamples = 0;

	for (int j = 0; j <  BRDF_SAMPLING_RES_THETA_H ; j++) {
		double phi_h = 0.0;
		double theta_h =  j * j * M_PI * 0.5 / (BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_H);
		double d_theta_h = 2. * sqrt(theta_h) * sqrt(0.5 * M_PI) / BRDF_SAMPLING_RES_THETA_H;
		for (int i = 0; i < BRDF_SAMPLING_RES_THETA_D ; i++) {
			double theta_d = i * M_PI * 0.5 / BRDF_SAMPLING_RES_THETA_D;
			double d_theta_d = M_PI * 0.5 / BRDF_SAMPLING_RES_THETA_D;
			for (int kn = 0; kn < BRDF_SAMPLING_RES_PHI_D / 2 ; kn++) {
				double phi_d = kn * M_PI * 2. / BRDF_SAMPLING_RES_PHI_D ;
				double d_phi_d =  M_PI * 2. / BRDF_SAMPLING_RES_PHI_D;
				double theta_i, theta_o, phi_i, phi_o; 
				MERLBRDF::half_diff_coords_to_std_coords(theta_h, phi_h, theta_d, phi_d, theta_i, phi_i, theta_o, phi_o);
				if (theta_i >= 0.5 * M_PI) continue;
				if (theta_o >= 0.5 * M_PI) continue;
				int phi_diff_index = kn;
				int theta_diff_index = i;
				int theta_half_index = j; 
				int ind = phi_diff_index +
					theta_diff_index * BRDF_SAMPLING_RES_PHI_D / 2 +
					theta_half_index * BRDF_SAMPLING_RES_PHI_D / 2 * BRDF_SAMPLING_RES_THETA_D;

				double wi = cos(theta_i) * cos(theta_o); 
				double wv = sqrt(8 * sin(theta_d) * sin(theta_d) * (cos(theta_d) * cos(theta_d) +  sin(theta_d) * sin(theta_d) * cos(phi_d) * cos(phi_d))) * sqrt(theta_h) * d_theta_h * d_theta_d * d_phi_d; 

				weights[numSamples] = wv; // * wi;
				totalWeight += weights[numSamples];
				for (int channel = 0; channel < 3; channel++) {
					values[numSamples + skipReading[channel]] = brdfMesured[ind + skipReading[channel]] * scaleReading[channel];
				}
				numSamples++;
			}
		}
	}
	if (numSamples > 0) {
		for (int channel = 0; channel < 3; channel++) {
			std::iota(idx, idx + numSamples - 1, 0);
			startingPoint = skipReading[channel];
			std::sort(idx, idx + numSamples - 1, compareIndices);
			// with C++11, this would work (less local variables)
			// std::sort(idx, idx + numSamples - 1, [values](int i1, int i2) {return values[skipReading[channel] + i1] < values[skipReading[channel] + i2];});

			double accWeight = 0;
			for (int i = 0; i < numSamples - 1; i++) {
				accWeight += weights[idx[i]];
				if (accWeight >= 0.5 * totalWeight) {
					median[channel] = values[skipReading[channel] + idx[i]];
					break;
				}
			}
			// std::cerr << "# " << values[skipReading[channel] + idx[0]] << " " << median[channel] << " " << values[skipReading[channel] + idx[numSamples - 2]] << std::endl;
		}
	}

	delete [] values;
	delete [] weights;
	delete [] idx;
}
