/**
 * Copyright (C) 2017 INRIA  http://www.inria.fr
 * Copyright (C) 2017 CNRS   http://www.cnrs.fr
 *
 * Authors:
 *  Nicolas Holzschuch:  nicolas.holzschuch@inria.fr
 *  Romain  Pacanowski:  romain.pacanowski@institutoptique.fr 
 *
 * Permission to use, copy and modify this software and its documentation without
 * fee for educational, research and non-profit purposes, is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * 
 * To request permission to incorporate this software into commercial products contact
 * the authors above. 
 */

#include <iostream>
#include <math.h>
#define M_2PI 6.2831853072
#include <assert.h>

// Code for computing the renormalization constant sigma_rel2_over_sigma_s2.
// Nothing fancy, just a lot of computations with very small dx. 
// Compile with:  g++ computeRenormalization.cpp -lm 
// Execute with: ./a.out > ctd_renormalization.h
//
// We cache only the integral part.
//
inline double integral_sigma_rel2(double lambda2_b2, double sin_theta_i, double c) {
	double liminf = 1 - sin_theta_i; 
	double limsup = 1 + sin_theta_i; 

	if (lambda2_b2 == 0) lambda2_b2 = 1e-7;
	double dv = 0.001 * M_PI/180;  
	double result = 0.0; 
	for (double v = liminf; v < limsup; v += dv) {
		double r2 = v * v ;
		double r = v ;
		double cos_thetaM = (1 - r2 - sin_theta_i * sin_theta_i) / (2 * r * sin_theta_i);
		if (cos_thetaM > 1) cos_thetaM = 1;
		if (cos_thetaM < -1) cos_thetaM = -1;
		result += 2 * (M_PI - acos(cos_thetaM)) * v * dv * pow(1 + v * v / lambda2_b2, -0.5*(c+1.));
		assert(!isnan(result)); 
	}
    assert(!isnan(result)); 
    assert(result >= 0); 
	return result;
}




int main() {
	int numSamples = 100;
	double c_min = 1.1;
	double c_max = 10.0;
	double delta_c = (c_max - c_min) / numSamples;
	double d_sin = 1.0 / numSamples; 

	std::cout << "\t double normalizationConstant[100][100][100] = {";
	for (int i = 0; i < numSamples; i++) {
		double l_b = 0.1 * (double)(i + 1) / (numSamples - i);
		double i_b = (numSamples + 1) * (10 * l_b / (1 + 10 * l_b)) - 1;
		
		std::cout << "{";
		for (double j = 0; j < numSamples; j++) {
			double c = (double)(numSamples + 1)/ (numSamples - j);
			double j_c = (numSamples + 1) * (c - 1.)/c - 1;
			std::cout << "{";
			for (double sti = 0; sti < 1.0; sti += d_sin) {
				double norm = integral_sigma_rel2(l_b * l_b, sti, c) ; // centerValue;
				std::cout << norm << ", " ;
			}
			std::cout << "}," << std::endl;
		}
		std::cout << "}," << std::endl;
	}
	std::cout << "};" << std::endl;
}
