/**
 * Copyright (C) 2017 INRIA  http://www.inria.fr
 * Copyright (C) 2017 CNRS   http://www.cnrs.fr
 *
 * Authors:
 *  Nicolas Holzschuch:  nicolas.holzschuch@inria.fr
 *  Romain  Pacanowski:  romain.pacanowski@institutoptique.fr 
 *
 * Permission to use, copy and modify this software and its documentation without
 * fee for educational, research and non-profit purposes, is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * 
 * To request permission to incorporate this software into commercial products contact
 * the authors above. 
 */

#include <iostream>
#include <fstream>
#include <assert.h>
#include <math.h>

// Code for computing shadowing term for the Exponential Power Distribution. 
// Two steps: first, compute integral of P22 (see integralP2_dataGen.cpp) and store into temporary file.
// Second, compute integral of the integral (this file), and store into header file. 
//
// Compile with:  g++ integralP2_G1.cpp -lm 
// Execute with: ./a.out > precomputed_G1.h


double P2[100][1000]; // P2 values, indexed by kp = 5./p and ks = exp(-u^(1/20))
double G1[100][1000]; // G1 values, indexed by kp = 5./p and ks = exp(-(b tan theta)^(1/20))
double dks = 0.001;

double G1_exp(double b_tan_theta, double p)
{
  // Need to compute an integral
  double ustart = 1./b_tan_theta;
  double exponent = 20; 
  double ks = exp(-pow(ustart, 1./exponent));
  int kks = ceil(1000 - 1000 * ks);
  ks = 1 - kks/1000.;
  int pmin = floor(5./p);
  int pmax = ceil(5./p);
  double alpha_p = 5./p - pmin;
  

  double oldu = ustart;
  double lambda = 0;
  for (;kks < 1000; kks++) {
    double ks = 1 - kks/1000.;
    double nextks =  1 - (kks+1)/1000.;
    double u = pow(log(1/ks), exponent);
    double nextu = pow(log(1/nextks), exponent);
    if (isinf(nextu)) nextu = u;    
    double deltau = 0.5 * (nextu - oldu);
    if (isnan(deltau)) deltau = 0;
    // std::cout << u << " " << ks/dks << " " << kks << std::endl ;
    double value = alpha_p * P2[pmax-1][kks] + (1 - alpha_p) * P2[pmin-1][kks];
    lambda += deltau * (b_tan_theta * u - 1) * value;
    // std::cout << u << " " << ks/dks << " " << kks << " " << value << " " << deltau << " " << lambda << std::endl ;
    oldu = u;
  }
  return 1./(1. + lambda);
}


int main() {
  double dkt = 0.0001;

  std::string G1Data("shadowingData");
  std::fstream fs;
  fs.open (G1Data.c_str(), std::fstream::in);

  // Reading data into memory
  for (int kp = 100; kp > 0; kp--) {
    double p = 5./kp;
    double integral2 = 0;
    double oldu = 0;
    double exponent = 20; // 2/(p);
    for (int kks = 0; kks < 1000; kks++) {
      double ks = 1 - kks/1000.;
      double nextks =  1 - (kks+1)/1000.;
      double u = pow(log(1/ks), exponent);
      double nextu = pow(log(1/nextks), exponent);
      if (isinf(nextu)) nextu = u;
      double deltau = (nextu - oldu);
      if (isnan(deltau)) deltau = 0;
      double value;
      fs >> value; 
      P2[kp-1][kks] = value;
      integral2 += deltau * value;
      // std::cout << u << " " << p << " " << value << " " << kks << " " << kp << " " << integral2 << " " << deltau << std::endl;
      oldu = u;
    }
    // std::cout << std::endl;
    // std::cout << "# p = " << p << " integral = " << integral2 << " " << kp << std::endl;
  }
  fs.close();

  std::cout << "double precomputedG1[100][1000] = {\n{";
  // Now compute G1(b * tan(theta)) :
  for (int kp = 100; kp > 0; kp--) {
    double p = 5./kp;
    double exponent = 20; // 2/(p);
    for (int kks0 = 0; kks0 < 1000; kks0++) {
      double ks0 = 1 - kks0/1000.;
      double u0 = pow(log(1/ks0), exponent);
      double oldu = u0;
      double Lambda = 0;
      double b_tan_theta = 1./u0;
      for (int kks = kks0; kks < 1000; kks++) {
        double ks = 1 - kks/1000.;
        double nextks =  1 - (kks+1)/1000.;
        double u = pow(log(1/ks), exponent);
        double nextu = pow(log(1/nextks), exponent);
        if (isinf(nextu)) nextu = u;
        double deltau = 0.5 * (nextu - oldu);
        if (isnan(deltau)) deltau = 0;
        Lambda += (b_tan_theta * u - 1) * P2[kp-1][kks] * deltau;
        // std::cout << u << " " << p << " " << value << " " << kks << " " << kp << " " << integral2 << " " << deltau << std::endl;
        oldu = u;
      }
      // std::cout << p << " " << u0 << " " << 1./(1. + Lambda) << " " <<  G1_exp(b_tan_theta, p) << " " << Lambda << std::endl;
      G1[kp-1][kks0] = 1./(1. + Lambda);
      if (isnan(G1[kp-1][kks0])) G1[kp-1][kks0] = 0.0;
      // this line to generate the precomputed_G1.h
      std::cout << G1[kp-1][kks0] << ", " ;
      // std::cout << u0 << " " << p << " " << G1[kp-1][kks0] << std::endl;
    }
   if (kp > 1) std::cout << "},\n {";
   else std::cout << "}};" << std::endl;
  }
}
